@@ -2,6 +2,7 @@
 
 pub mod attributes;
 
+#[cfg(feature = "encoding_rs")]
 use encoding_rs::Encoding;
 use std::borrow::Cow;
 use std::io::BufRead;
@@ -26,7 +27,7 @@ use memchr;
 /// [`local_name`]: #method.local_name
 /// [`unescaped`]: #method.unescaped
 /// [`attributes`]: #method.attributes
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct BytesStart<'a> {
     /// content of the element, before any utf8 conversion
     buf: Cow<'a, [u8]>,
@@ -36,6 +37,10 @@ pub struct BytesStart<'a> {
 
 impl<'a> BytesStart<'a> {
     /// Creates a new `BytesStart` from the given content (name + attributes).
+    ///
+    /// # Warning
+    ///
+    /// `&content[..name_len]` is not checked to be a valid name
     #[inline]
     pub fn borrowed(content: &'a [u8], name_len: usize) -> BytesStart<'a> {
         BytesStart {
@@ -45,12 +50,13 @@ impl<'a> BytesStart<'a> {
     }
 
     /// Creates a new `BytesStart` from the given name.
+    ///
+    /// # Warning
+    ///
+    /// `&content` is not checked to be a valid name
     #[inline]
     pub fn borrowed_name(name: &'a [u8]) -> BytesStart<'a> {
-        BytesStart {
-            name_len: name.len(),
-            buf: Cow::Borrowed(name),
-        }
+        Self::borrowed(name, name.len())
     }
 
     /// Creates a new `BytesStart` from the given content (name + attributes)
@@ -78,10 +84,7 @@ impl<'a> BytesStart<'a> {
 
     /// Converts the event into an owned event.
     pub fn into_owned(self) -> BytesStart<'static> {
-        BytesStart {
-            buf: Cow::Owned(self.buf.into_owned()),
-            name_len: self.name_len,
-        }
+        Self::owned(self.buf.into_owned(), self.name_len)
     }
 
     /// Consumes `self` and yield a new `BytesStart` with additional attributes from an iterator.
@@ -99,6 +102,7 @@ impl<'a> BytesStart<'a> {
     }
 
     /// Gets the undecoded raw tag name as a `&[u8]`.
+    #[inline]
     pub fn name(&self) -> &[u8] {
         &self.buf[..self.name_len]
     }
@@ -116,6 +120,7 @@ impl<'a> BytesStart<'a> {
     ///
     /// XML escape sequences like "`&lt;`" will be replaced by their unescaped characters like
     /// "`<`".
+    #[inline]
     pub fn unescaped(&self) -> Result<Cow<[u8]>> {
         unescape(&*self.buf).map_err(Error::EscapeError)
     }
@@ -156,10 +161,29 @@ impl<'a> BytesStart<'a> {
     ///
     /// [`unescaped()`]: #method.unescaped
     /// [`Reader::decode()`]: ../reader/struct.Reader.html#method.decode
+    #[cfg(feature = "encoding_rs")]
+    #[inline]
     pub fn unescape_and_decode<B: BufRead>(&self, reader: &Reader<B>) -> Result<String> {
         self.unescaped().map(|e| reader.decode(&*e).into_owned())
     }
 
+    /// Returns the unescaped and decoded string value.
+    ///
+    /// This allocates a `String` in all cases. For performance reasons it might be a better idea to
+    /// instead use one of:
+    ///
+    /// * [`unescaped()`], as it doesn't allocate when no escape sequences are used.
+    /// * [`Reader::decode()`], as it only allocates when the decoding can't be performed otherwise.
+    ///
+    /// [`unescaped()`]: #method.unescaped
+    /// [`Reader::decode()`]: ../reader/struct.Reader.html#method.decode
+    #[cfg(not(feature = "encoding_rs"))]
+    #[inline]
+    pub fn unescape_and_decode<B: BufRead>(&self, reader: &Reader<B>) -> Result<String> {
+        self.unescaped()
+            .and_then(|e| reader.decode(&*e).map(|s| s.to_owned()))
+    }
+
     /// Adds an attribute to this element.
     pub fn push_attribute<'b, A: Into<Attribute<'b>>>(&mut self, attr: A) {
         let a = attr.into();
@@ -172,11 +196,13 @@ impl<'a> BytesStart<'a> {
     }
 
     /// Edit the name of the BytesStart in-place
+    ///
+    /// # Warning
+    ///
+    /// `name` is not checked to be a valid name
     pub fn set_name(&mut self, name: &[u8]) -> &mut BytesStart<'a> {
-        {
-            let bytes = self.buf.to_mut();
-            bytes.splice(..self.name_len, name.iter().cloned());
-        }
+        let bytes = self.buf.to_mut();
+        bytes.splice(..self.name_len, name.iter().cloned());
         self.name_len = name.len();
         self
     }
@@ -188,6 +214,16 @@ impl<'a> BytesStart<'a> {
     }
 }
 
+impl<'a> std::fmt::Debug for BytesStart<'a> {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        use crate::utils::write_byte_string;
+
+        write!(f, "BytesStart {{ buf: ")?;
+        write_byte_string(f, &self.buf)?;
+        write!(f, ", name_len: {} }}", self.name_len)
+    }
+}
+
 /// An XML declaration (`Event::Decl`).
 ///
 /// [W3C XML 1.1 Prolog and Document Type Declaration](http://w3.org/TR/xml11/#sec-prolog-dtd)
@@ -297,6 +333,7 @@ impl<'a> BytesDecl<'a> {
     }
 
     /// Gets the decoder struct
+    #[cfg(feature = "encoding_rs")]
     pub fn encoder(&self) -> Option<&'static Encoding> {
         self.encoding()
             .and_then(|e| e.ok())
@@ -312,7 +349,7 @@ impl<'a> BytesDecl<'a> {
 }
 
 /// A struct to manage `Event::End` events
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct BytesEnd<'a> {
     name: Cow<'a, [u8]>,
 }
@@ -360,8 +397,18 @@ impl<'a> BytesEnd<'a> {
     }
 }
 
+impl<'a> std::fmt::Debug for BytesEnd<'a> {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        use crate::utils::write_byte_string;
+
+        write!(f, "BytesEnd {{ name: ")?;
+        write_byte_string(f, &self.name)?;
+        write!(f, " }}")
+    }
+}
+
 /// Data from various events (most notably, `Event::Text`).
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct BytesText<'a> {
     // Invariant: The content is always escaped.
     content: Cow<'a, [u8]>,
@@ -424,16 +471,39 @@ impl<'a> BytesText<'a> {
     /// it might be wiser to manually use
     /// 1. BytesText::unescaped()
     /// 2. Reader::decode(...)
+    #[cfg(feature = "encoding_rs")]
     pub fn unescape_and_decode<B: BufRead>(&self, reader: &Reader<B>) -> Result<String> {
         self.unescaped().map(|e| reader.decode(&*e).into_owned())
     }
 
+    /// helper method to unescape then decode self using the reader encoding
+    ///
+    /// for performance reasons (could avoid allocating a `String`),
+    /// it might be wiser to manually use
+    /// 1. BytesText::unescaped()
+    /// 2. Reader::decode(...)
+    #[cfg(not(feature = "encoding_rs"))]
+    pub fn unescape_and_decode<B: BufRead>(&self, reader: &Reader<B>) -> Result<String> {
+        self.unescaped()
+            .and_then(|e| reader.decode(&*e).map(|s| s.to_owned()))
+    }
+
     /// Gets escaped content.
     pub fn escaped(&self) -> &[u8] {
         self.content.as_ref()
     }
 }
 
+impl<'a> std::fmt::Debug for BytesText<'a> {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        use crate::utils::write_byte_string;
+
+        write!(f, "BytesText {{ content: ")?;
+        write_byte_string(f, &self.content)?;
+        write!(f, " }}")
+    }
+}
+
 /// Event emitted by [`Reader::read_event`].
 ///
 /// [`Reader::read_event`]: ../reader/struct.Reader.html#method.read_event
